/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.algo;

import java.io.PrintStream;
import java.util.Map.Entry;

import at.jku.risc.stout.hoau.data.EquationSystem;
import at.jku.risc.stout.hoau.data.TermNode;
import at.jku.risc.stout.hoau.data.atom.BoundVariable;
import at.jku.risc.stout.hoau.data.atom.Variable;
import at.jku.risc.stout.hoau.util.ControlledException;

/**
 * This class encapsulates the rule based system {@linkplain AntiUnifySystem}
 * and allows to justify the result by back substitution of the generalization
 * with the reported differences inside the computed store and testing alpha
 * equivalence with also allowing eta expansion.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnify {
	private DebugLevel debugLevel;
	private EquationSystem<AntiUnifyProblem> eqSysOrig;
	private AntiUnifySystem hoauSys;
	private int maxNormalizeSteps;

	/**
	 * Creates and initializes the encapsulated rule based system
	 * {@linkplain AntiUnifySystem} with the given equation system and an empty
	 * {@linkplain Substitution}.
	 */
	public AntiUnify(EquationSystem<AntiUnifyProblem> eq,
			int maxNormalizeSteps, DebugLevel debugLevel)
			throws ControlledException {
		this.debugLevel = debugLevel;
		this.eqSysOrig = eq;
		this.maxNormalizeSteps = maxNormalizeSteps;
		this.hoauSys = new AntiUnifySystem(eq.clone(), new Substitution(
				maxNormalizeSteps));
	}

	/**
	 * Calls the rule based system {@linkplain AntiUnifySystem} and tries to
	 * justify the computed result if the second argument of this method is
	 * true.
	 */
	public void antiUnify(boolean justify, PrintStream out)
			throws ControlledException {
		boolean originName = BoundVariable.PRINT_ORIGIN_NAME;
		if (debugLevel == DebugLevel.PROGRESSORIGIN)
			BoundVariable.PRINT_ORIGIN_NAME = true;
		try {
			antiUnifyIntern(justify, out);
		} finally {
			BoundVariable.PRINT_ORIGIN_NAME = originName;
		}
	}

	private void antiUnifyIntern(boolean justify, PrintStream out)
			throws ControlledException {
		hoauSys.compute(debugLevel, out);
		if (debugLevel == DebugLevel.SIMPLE) {
			printResult(out);
		} else if (debugLevel != DebugLevel.SILENT) {
			out.println("-----------");
			printResult(out);
			out.println("-----------");
			out.println("  Sigma: " + hoauSys.getSigma());
			out.println("  Store: " + hoauSys.getStore());
			out.println();
		}
		if (justify) {
			if (debugLevel == DebugLevel.PROGRESS)
				out.println("Start justification.");
			for (AntiUnifyProblem aup : eqSysOrig)
				justifyAup(out, aup);
		}
		for (Entry<Variable, TermNode> solved : hoauSys.getSigma().getMapping()
				.entrySet())
			callback(hoauSys, solved.getKey());
	}

	/**
	 * This callback function will be invoked for every found generalization. By
	 * default it does nothing. For instance, one can use this callback function
	 * to collect all the results or to filter out some interesting results.<br>
	 * In combination with a matching algorithm, one can obtain the minimal
	 * complete set of generalizations by using this callback function.
	 */
	public void callback(AntiUnifySystem result, Variable generalizationVar) {
	}

	private void justifyAup(PrintStream out, AntiUnifyProblem aup)
			throws ControlledException {
		TermNode result = hoauSys.getSigma().get(aup.getGeneralizationVar());
		if (result == null)
			throw new JustificationException("No result found for " + aup);
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("Generating two lambda-terms from the computed generalization");
			out.println("by generating two substitutions from the store and applying them.");
		}
		TermNode lRes = result;
		TermNode rRes = result.clone();
		for (int j = hoauSys.getStore().size() - 1; j >= 0; j--) {
			AntiUnifyProblem backSubsti = hoauSys.getStore().get(j);
			TermNode lSigma = backSubsti.bindAll(backSubsti.getLeft());
			TermNode rSigma = backSubsti.bindAll(backSubsti.getRight());
			lRes.substitute(backSubsti.getGeneralizationVar(), lSigma);
			rRes.substitute(backSubsti.getGeneralizationVar(), rSigma);
			lRes = lRes.reduce(maxNormalizeSteps);
			rRes = rRes.reduce(maxNormalizeSteps);
		}
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("Testing alpha equivalence of the generated terms with the input terms.");
		}
		TermNode lOri = aup.getLeft().clone();
		TermNode rOri = aup.getRight().clone();
		if (!lRes.equals(lOri) || !rRes.equals(rOri)) {
			out.println("ERROR INPUT: " + lOri
					+ AntiUnifyProblem.PRINT_EQ_SEPARATOR + rOri);
			out.println("ERROR OUTPUT: " + lRes
					+ AntiUnifyProblem.PRINT_EQ_SEPARATOR + rRes);
			throw new JustificationException(
					"The computed result is not a valid generalization");
		} else if (debugLevel != DebugLevel.SILENT
				&& debugLevel != DebugLevel.SIMPLE) {
			out.println("Alpha equivalence ok: " + lRes
					+ AntiUnifyProblem.PRINT_EQ_SEPARATOR + rRes);
		}
	}

	private void printResult(PrintStream out) {
		out.println(" Result: " + hoauSys);
		BoundVariable.PRINT_ORIGIN_NAME = !BoundVariable.PRINT_ORIGIN_NAME;
		out.println(" Origin: " + hoauSys);
		BoundVariable.PRINT_ORIGIN_NAME = !BoundVariable.PRINT_ORIGIN_NAME;
	}

	/**
	 * Returns the original equation system.
	 */
	public EquationSystem<AntiUnifyProblem> getEqSysOrig() {
		return eqSysOrig;
	}

	/**
	 * Returns the entire {@linkplain AntiUnifySystem} which is encapsulated by
	 * this class
	 */
	public AntiUnifySystem getResult() {
		return hoauSys;
	}
}
