/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.algo;

import java.io.PrintStream;
import java.util.List;
import java.util.Map;

import at.jku.risc.stout.hoau.data.EquationSystem;
import at.jku.risc.stout.hoau.data.Hedge;
import at.jku.risc.stout.hoau.data.NotNormalizableException;
import at.jku.risc.stout.hoau.data.TermNode;
import at.jku.risc.stout.hoau.data.atom.BoundVariable;
import at.jku.risc.stout.hoau.data.atom.Lambda;
import at.jku.risc.stout.hoau.data.atom.TermAtom;
import at.jku.risc.stout.hoau.data.atom.Variable;
import at.jku.risc.stout.hoau.util.DataStructureFactory;

/**
 * This class represents a rule based system for (a variant of) higher-order
 * anti-unification.<br>
 * The algorithm P is described in the paper:<br>
 * <a href="http://www.risc.jku.at/home/abaumgar/">Alexander Baumgartner</a>, <a
 * href="http://www.risc.jku.at/people/tkutsia/">Temur Kutsia</a>, <a
 * href="http://www.iiia.csic.es/~levy/">Jordi Levy</a> and <a
 * href="http://ima.udg.edu/~villaret/">Mateu Villaret</a>, A Variant of
 * Higher-Order Anti-Unification
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifySystem {
	private PermEquivSystem mr = new PermEquivSystem();
	private EquationSystem<AntiUnifyProblem> problemSet;
	private List<AntiUnifyProblem> store = DataStructureFactory.$.newList();
	private Substitution sigma;

	/**
	 * Creates and initializes the rule based system.
	 */
	public AntiUnifySystem(EquationSystem<AntiUnifyProblem> problemSet,
			Substitution sigma) throws NotNormalizableException {
		this.problemSet = problemSet;
		this.sigma = sigma;
		for (AntiUnifyProblem aup : problemSet) {
			sigma.put(aup.getGeneralizationVar(),
					new TermNode(aup.getGeneralizationVar(), null));
		}
	}

	/**
	 * Computes the result silently.
	 */
	public void compute() throws NotNormalizableException {
		compute(DebugLevel.SILENT, null);
	}

	/**
	 * Computes the result and prints progress information to
	 * {@linkplain PrintStream} according to {@linkplain DebugLevel}.
	 */
	public void compute(DebugLevel debugLevel, PrintStream debugOut)
			throws NotNormalizableException {
		while (!problemSet.isEmpty()) {
			debug(debugLevel, debugOut);

			AntiUnifyProblem eq = problemSet.popFirst();
			TermNode left = eq.getLeft();
			TermNode right = eq.getRight();
			TermAtom lAtom = left.getAtom();
			TermAtom rAtom = right.getAtom();
			Hedge lHedge = left.getHedge();
			Hedge rHedge = right.getHedge();

			// Abstraction
			if (lAtom instanceof Lambda || rAtom instanceof Lambda) {
				if (!(lAtom instanceof Lambda))
					left.etaExpand(((Lambda) rAtom).getBoundVar());
				else if (!(rAtom instanceof Lambda))
					right.etaExpand(((Lambda) lAtom).getBoundVar());
				eq.abstrLambda();
				problemSet.add(eq);
				sigma.composeInRange(eq.getGeneralizationVar(),
						eq.bindAll(eq.createHO_Fnc()));
				debug("ABS", debugLevel, debugOut);
			}
			// Decomposition
			else if (lAtom.equals(rAtom) && lHedge.size() == rHedge.size()) {
				Hedge substiHedge = lHedge.newInstance(); // Hedge of same type
				TermNode substi = new TermNode(lAtom, substiHedge);
				for (int i = 0, n = lHedge.size(); i < n; i++) {
					AntiUnifyProblem newEq = problemSet.newEquation();
					newEq.setLeft(lHedge.get(i));
					newEq.setRight(rHedge.get(i));
					newEq.copyAbstr(eq);
					problemSet.add(newEq);
					substiHedge.add(newEq.createHO_Fnc());
				}
				sigma.composeInRange(eq.getGeneralizationVar(),
						eq.bindAll(substi));
				debug("DEC", debugLevel, debugOut);
			}
			// Solve
			else {
				AntiUnifyProblem newEq = problemSet.newEquation();
				newEq.setLeft(left);
				newEq.setRight(right);
				for (int i = 0, n = eq.abstrSize(); i < n; i++) {
					BoundVariable var = eq.abstrItem(i);
					if (left.hasFreeVar(var) || right.hasFreeVar(var))
						newEq.addAbstrVar(var);
				}
				store.add(newEq);
				sigma.composeInRange(eq.getGeneralizationVar(),
						eq.bindAll(newEq.createHO_Fnc()));
				debug("SOL", debugLevel, debugOut);
			}
		}
		debug(debugLevel, debugOut);
		// Merge
		for (int i = store.size() - 2; i >= 0; i--) {
			AntiUnifyProblem eq1 = store.get(i);
			for (int j = store.size() - 1; j > i; j--) {
				AntiUnifyProblem eq2 = store.get(j);
				if (eq1.abstrSize() == eq2.abstrSize()) {
					mr.reset();
					mr.start(eq1.abstrSet(), eq2.abstrSet());
					mr.addEquation(eq1.getLeft(), eq2.getLeft());
					mr.addEquation(eq1.getRight(), eq2.getRight());
					Map<Variable, Variable> pi;
					pi = mr.compute(debugLevel, debugOut);
					if (pi != null) {
						sigma.composeInRange(eq2.getGeneralizationVar(),
								eq2.bindAll(eq1.createHO_Fnc(pi)));
						store.remove(j);
						debug("MER", debugLevel, debugOut);
						debug(debugLevel, debugOut);
					}
				}
			}
		}
	}

	private void debug(DebugLevel debugLevel, PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS
				|| debugLevel == DebugLevel.PROGRESSORIGIN) {
			out.println(" Problem: " + problemSet);
			out.println("   Store: " + store);
			out.println("   Sigma: " + sigma);
			out.println();
		}
	}

	private void debug(String rule, DebugLevel debugLevel, PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS
				|| debugLevel == DebugLevel.PROGRESSORIGIN) {
			out.println(rule + " ==> ");
		}
	}

	/**
	 * Returns the problem set which is empty after {@linkplain #compute()}
	 * succeeded.
	 */
	public EquationSystem<AntiUnifyProblem> getProblemSet() {
		return problemSet;
	}

	/**
	 * Returns the store where all the disagreements are stored during the
	 * computation. It contains all the information about how to instantiate the
	 * computed generalization to obtain the original terms back. Only contains
	 * information after {@linkplain #compute()} succeeded.
	 */
	public List<AntiUnifyProblem> getStore() {
		return store;
	}

	/**
	 * The substitution sigma contains the generalization after
	 * {@linkplain #compute()} succeeded.
	 */
	public Substitution getSigma() {
		return sigma;
	}

	@Override
	public String toString() {
		return sigma.toRanString();
	}
}
