/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.algo;

import java.io.PrintStream;
import java.util.Map;
import java.util.Set;

import at.jku.risc.stout.hoau.data.EquationSystem;
import at.jku.risc.stout.hoau.data.atom.Variable;
import at.jku.risc.stout.hoau.util.ControlledException;

/**
 * This class encapsulates the rule based system {@linkplain PermEquivSystem}.
 * It computes a permutation such that the equations in the given
 * {@linkplain EquationSystem} are all alpha equivalent, if it exists. Otherwise
 * it returns null.
 * 
 * @author Alexander Baumgartner
 */
public class PermEquiv {
	private EquationSystem<PermEquivProblem> eqSys;
	private Set<Variable> dom;
	private Set<Variable> ran;

	/**
	 * Creates and initializes the encapsulated rule based system
	 * {@linkplain PermEquivSystem} with the given equation system, domain and
	 * range.
	 */
	public PermEquiv(EquationSystem<PermEquivProblem> eqSys, Set<Variable> dom,
			Set<Variable> ran) {
		this.eqSys = eqSys;
		this.dom = dom;
		this.ran = ran;
	}

	/**
	 * Calls the rule based system {@linkplain PermEquivSystem} and returns the
	 * result.
	 */
	public Map<Variable, Variable> compute(DebugLevel debugLevel,
			PrintStream out) throws ControlledException {
		PermEquivSystem perm = new PermEquivSystem(eqSys.clone());
		perm.start(dom, ran);
		Map<Variable, Variable> pi = perm.compute(debugLevel, out);

		String result = pi == null ? "No mapping found." : pi.toString();
		if (debugLevel != DebugLevel.SILENT) {
			out.println("-----------");
			out.println(" Result: " + result);
			out.println("-----------");
			out.println();
		}
		if (debugLevel != DebugLevel.SILENT && debugLevel != DebugLevel.SIMPLE)
			perm.printState(out);

		return pi;
	}

	public EquationSystem<PermEquivProblem> getEqSys() {
		return eqSys;
	}

	public Set<Variable> getDom() {
		return dom;
	}

	public Set<Variable> getRan() {
		return ran;
	}
}
