/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.data;

import java.util.Deque;
import java.util.Iterator;

import at.jku.risc.stout.hoau.util.DataStructureFactory;

/**
 * A system of {@linkplain Equation}s
 * 
 * @author Alexander Baumgartner
 */
public abstract class EquationSystem<T extends Equation> implements Cloneable,
		Iterable<T> {
	private Deque<T> equations = DataStructureFactory.$.newDeque();
	/**
	 * The separator is used to separate the equations of this system.
	 */
	public static String EQ_SEPARATOR = ";  ";
	/**
	 * The output will look like "{@linkplain #EQ_PREFIX1}#i#
	 * {@linkplain #EQ_PREFIX2}" where #i# is the index of an equation.
	 */
	public static String EQ_PREFIX1 = "EQ";
	/**
	 * If {@linkplain #EQ_PREFIX2} is null then the output of the index will
	 * also be omitted.
	 */
	public static String EQ_PREFIX2 = ": ";

	/**
	 * You have to implement this method, so that an algorithm is able to
	 * instantiate new {@linkplain Equation}s of arbitrary types.
	 */
	public abstract T newEquation();

	/**
	 * Instantiates an equation of the specified type and initializes it with
	 * the given {@linkplain TermNode}s.
	 */
	public T newEquation(TermNode left, TermNode right) {
		T eq = newEquation();
		eq.setLeft(left);
		eq.setRight(right);
		return eq;
	}

	/**
	 * Returns the number of equations in this system.
	 */
	public int size() {
		return equations.size();
	}

	/**
	 * Tests whether the system of equations is empty.
	 */
	public boolean isEmpty() {
		return equations.isEmpty();
	}

	/**
	 * Adds a new equation to this system.
	 */
	public void add(T equation) {
		equations.add(equation);
	}

	/**
	 * Retrieves and removes the first element of this system of equations or
	 * returns null if it is empty.
	 */
	public T popFirst() {
		return equations.pollFirst();
	}

	/**
	 * Retrieves and removes the last element of this system of equations or
	 * returns null if it is empty.
	 */
	public T popLast() {
		return equations.pollLast();
	}

	/**
	 * Retrieves the first element of this system of equations or returns null
	 * if it is empty.
	 */
	public T getFirst() {
		return equations.peekFirst();
	}

	/**
	 * Retrieves the last element of this system of equations or returns null if
	 * it is empty.
	 */
	public T getLast() {
		return equations.peekLast();
	}

	/**
	 * Removes all the equations of this system.
	 */
	public void clear() {
		equations.clear();
	}

	@Override
	public Iterator<T> iterator() {
		return equations.iterator();
	}

	@Override
	@SuppressWarnings("unchecked")
	public EquationSystem<T> clone() {
		try {
			EquationSystem<T> ret = (EquationSystem<T>) super.clone();
			ret.equations = DataStructureFactory.$.newDeque(equations.size());
			for (T elem : equations)
				ret.equations.add((T) elem.clone());
			return ret;
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		int i = 0;
		for (T elem : equations) {
			if (EQ_SEPARATOR != null && i != 0)
				sb.append(EQ_SEPARATOR);
			if (EQ_PREFIX1 != null)
				sb.append(EQ_PREFIX1);
			if (EQ_PREFIX2 != null)
				sb.append(i + 1).append(EQ_PREFIX2);
			sb.append(elem);
			i++;
		}
		return sb.toString();
	}
}
