/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.data;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;

import at.jku.risc.stout.hoau.util.ControlledException;

/**
 * Tiny term parser with the following rules: (The described rules use the
 * default configuration. You can change the special symbols and/or the variable
 * range.)
 * <ul>
 * <li>The symbols A-Z, a-z, 1-9 are allowed for variable names and function
 * names. Names can be of any length.<br>
 * <li>If the first letter of the name is within [u,z] or [U,Z], then it is a
 * free variable.<br>
 * <li>The backslash "\" is used for the lambda symbol. The name followed by the
 * lambda symbol will become a bound variable.
 * <li>The colon ":" is used to declare the type. Type specification is optional
 * for every single symbol.
 * <li>The dash "-" is used as type constructor.
 * </ul>
 * 
 * @author Alexander Baumgartner
 */
public class InputParser {
	/**
	 * default = 'u' => A word starting with 'u-z' will become a variable.<br>
	 * {@linkplain InputParser#VARIABLE_START} has to be a lower case letter!
	 * Hedge variables are denoted by upper case letters.
	 */
	public static char VARIABLE_START = 'u';
	/**
	 * default = '\' => Any word after the '\' will become a bound variable.
	 */
	public static char LAMBDA_NAME = '\\';
	/**
	 * default = '-'.
	 */
	public static char TYPE_CONSTRUCTOR = '-';
	/**
	 * default = '('.
	 */
	public static char OPENING_PARENTHESIS = '(';
	/**
	 * default = ')'.
	 */
	public static char CLOSING_PARENTHESIS = ')';
	private int codePoint;
	private boolean forceType;
	private NodeFactory nodeFactory;

	/**
	 * Like {@linkplain #InputParser(NodeFactory, boolean)} with a new
	 * {@linkplain NodeFactory} instance and forceType set to false.
	 */
	public InputParser() {
		this(new NodeFactory(), false);
	}

	/**
	 * Constructor with the used {@linkplain NodeFactory}. If the second flag is
	 * set to true, then typing is mandatory.
	 */
	public InputParser(NodeFactory nodeFactory, boolean forceType) {
		this.nodeFactory = nodeFactory;
		this.forceType = forceType;
	}

	/**
	 * Parses a string and puts all the equations into the given equation
	 * system. Beta reduction is automatically called for every parsed term.
	 */
	public <T extends Equation> void parseEquationSystem(String in,
			EquationSystem<T> eqSys, int maxReduce) throws IOException,
			ControlledException {
		String[] eqList = in.trim().split("\\s*;\\s*");
		for (String eq : eqList) {
			if (eq.length() > 0) {
				String[] leftRight = eq.split("[\\s]*=[\\s?^=]*");
				TermNode left = parseSequence(leftRight[0]).reduce(maxReduce);
				TermNode right;
				if (leftRight.length == 1)
					right = new TermNode(null, null);
				else
					right = parseSequence(leftRight[1]).reduce(maxReduce);
				eqSys.add(eqSys.newEquation(left, right));
			}
		}
	}

	public <T extends Equation> void parseEquation(Reader in1, Reader in2,
			EquationSystem<T> eqSys, int maxReduce)
			throws NotNormalizableException, MalformedTermException,
			IOException {
		TermNode left = parseSequence(in1).reduce(maxReduce);
		TermNode right = parseSequence(in2).reduce(maxReduce);
		eqSys.add(eqSys.newEquation(left, right));
	}

	/**
	 * This method tries to find a suitable prefix for fresh free variables. Use
	 * this method carefully! (Or don't use it at all.)<br>
	 * The best way to go is to simply choose a special symbol for fresh
	 * variables which is guaranteed to never appear inside a term. (Just set
	 * {@linkplain NodeFactory#PREFIX_FreshFreeVar} and don't use this method!)
	 */
	public void tryFreeVarPrefix(String in, String... prefix) {
		String varPrefix = findVar(in, prefix);
		NodeFactory.PREFIX_FreshFreeVar = varPrefix == null ? "$" : varPrefix;
	}

	/**
	 * This method tries to find a suitable prefix for fresh bound variables.
	 * Use this method carefully! (Or don't use it at all.)<br>
	 * The best way to go is to simply choose a special symbol for fresh
	 * variables which is guaranteed to never appear inside a term. (Just set
	 * {@linkplain NodeFactory#PREFIX_FreshBoundVar} and don't use this method!)
	 */
	public void tryBoundVarPrefix(String in, String... prefix) {
		String varPrefix = findVar(in, prefix);
		NodeFactory.PREFIX_FreshBoundVar = varPrefix == null ? "#" : varPrefix;
	}

	private String findVar(String in, String... prefix) {
		for (String pr : prefix)
			if (!in.contains(pr))
				return pr;
		return null;
	}

	/**
	 * Use {@linkplain #parseSequence(Reader)} if possible! It performs better
	 * and saves space by reading directly from the data stream;)
	 */
	public TermNode parseSequence(String in) throws IOException,
			MalformedTermException {
		return parseSequence(new StringReader(in));
	}

	/**
	 * Parses a sequence of terms out of the given {@linkplain Reader} and
	 * returns it.
	 */
	public TermNode parseSequence(Reader in) throws IOException,
			MalformedTermException {
		codePoint = 0;
		nodeFactory.pushHedge(true);
		do {
			nodeFactory.addToHedge(parseIt(in));
		} while (codePoint != -1);
		Hedge h = nodeFactory.popHedge();
		return h.size() == 1 ? h.get(0) : new TermNode(null, h);
	}

	/**
	 * Use {@linkplain #parseTerm(Reader)} if possible! It performs better and
	 * saves space by reading directly from the data stream;)
	 */
	public TermNode parseTerm(String in) throws IOException,
			MalformedTermException {
		return parseTerm(new StringReader(in));
	}

	/**
	 * Parses a term out of the given {@linkplain Reader} and returns it.
	 */
	public TermNode parseTerm(Reader in) throws IOException,
			MalformedTermException {
		codePoint = 0;
		return parseIt(in);
	}

	private TermNode parseIt(Reader in) throws IOException,
			MalformedTermException {
		String name = parseTermName(in, false).toString();
		if (name.length() == 0) {
			if (codePoint == OPENING_PARENTHESIS) {
				nodeFactory.pushHedge(true);
				while (codePoint != CLOSING_PARENTHESIS
						&& next(in) != CLOSING_PARENTHESIS)
					nodeFactory.addToHedge(parseIt(in));
				Hedge hedge = nodeFactory.popHedge();
				next(in); // consume closing bracket
				return hedge.size() == 1 ? hedge.get(0) : new TermNode(null,
						hedge);
			} else {
				throw new MalformedTermException(
						"Missing parenthesis at the input term");
			}
		}
		char nameLetter1 = name.charAt(0);
		String type = null;
		if (codePoint == ':') { // type
			type = parseTermName(in, true).toString();
			if (type.length() == 0)
				throw new MalformedTermException(
						"Missing type at the input term");
		}
		if (nameLetter1 == LAMBDA_NAME) {
			name = name.substring(1);
			if (codePoint == ',') // short form
				codePoint = LAMBDA_NAME;
			if (codePoint == '.') // optional separator
				next(in);
		}
		if (codePoint == OPENING_PARENTHESIS) {
			if (nameLetter1 == LAMBDA_NAME)
				nodeFactory.pushLambda(name, type);
			nodeFactory.pushHedge(nameLetter1 == LAMBDA_NAME);
			while (codePoint != CLOSING_PARENTHESIS
					&& next(in) != CLOSING_PARENTHESIS)
				nodeFactory.addToHedge(parseIt(in));
			Hedge hedge = nodeFactory.popHedge();
			next(in); // consume closing bracket
			if (nameLetter1 == LAMBDA_NAME) {
				return check(nodeFactory.popLambda(hedge));
			} else if (isVariable(nameLetter1, name)) {
				return check(nodeFactory.createIndividualVar(name, type, hedge));
			} else {
				return check(nodeFactory.createFunction(name, type, hedge));
			}
		} else if (nameLetter1 == LAMBDA_NAME) {
			nodeFactory.pushLambda(name, type);
			nodeFactory.pushHedge(true);
			do {
				nodeFactory.addToHedge(parseIt(in));
				while (Character.isWhitespace(codePoint))
					next(in);
			} while (codePoint != -1
					&& (isNameChar(codePoint) || codePoint == OPENING_PARENTHESIS));
			Hedge hedge = nodeFactory.popHedge();
			return check(nodeFactory.popLambda(hedge));
		} else if (isVariable(nameLetter1, name)) {
			return check(nodeFactory.createIndividualVar(name, type));
		} else {
			return check(nodeFactory.createConstant(name, type));
		}
	}

	private boolean isVariable(char nameLetter1, String name) {
		return Character.toLowerCase(nameLetter1) >= VARIABLE_START
				|| nodeFactory.hasBoundVar(name);
	}

	private TermNode check(TermNode node) throws MalformedTermException {
		if (forceType && !node.isNullAtom() && node.getAtom().getType() == null) {
			throw new MalformedTermException("Type is missing for "
					+ node.getAtom().getOriginName());
		}
		return node;
	}

	private StringBuilder parseTermName(Reader in, boolean isType)
			throws IOException {
		StringBuilder name = new StringBuilder();
		while (codePoint != -1 && codePoint != OPENING_PARENTHESIS
				&& !isNameChar(codePoint))
			next(in);
		for (; (isType && codePoint == TYPE_CONSTRUCTOR)
				|| isNameChar(codePoint); next(in))
			name.appendCodePoint(codePoint);
		return name;
	}

	private int next(Reader in) throws IOException {
		if (Character.isWhitespace(codePoint))
			do
				codePoint = in.read();
			while (Character.isWhitespace(codePoint));
		else
			codePoint = in.read();
		return codePoint;
	}

	private boolean isNameChar(int codePoint) {
		return codePoint == '_' || codePoint == LAMBDA_NAME
				|| Character.isLetterOrDigit(codePoint);
	}
}
