/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.data;

import java.io.IOException;
import java.io.Writer;

import at.jku.risc.stout.hoau.data.atom.Lambda;
import at.jku.risc.stout.hoau.data.atom.TermAtom;
import at.jku.risc.stout.hoau.data.atom.Variable;

/**
 * A {@linkplain LambdaHedge} is a list of {@linkplain TermNode}s.
 * 
 * @author Alexander Baumgartner
 */
public class LambdaHedge extends Hedge {

	public static String PRINT_PARAM_START = "(";
	public static String PRINT_PARAM_SEPARATOR = " ";
	public static String PRINT_PARAM_END = ")";

	@Override
	public Hedge newInstance() {
		return new LambdaHedge();
	}

	@Override
	public Hedge reduce(int count) throws NotNormalizableException {
		Hedge hedge = super.reduce(count);
		for (int i = 0, n = hedge.size() - 1; i < n; i++) {
			TermNode firstNode = hedge.get(i);
			if (firstNode.getAtom() instanceof Lambda) {
				TermNode secondNode = hedge.get(i + 1);
				if (count <= 0)
					throw new NotNormalizableException(
							"Maximum amount of reduction steps exceeded");

				Lambda l = (Lambda) firstNode.getAtom();
				Hedge reduced = firstNode.getHedge();
				reduced.substitute(l.getBoundVar(), secondNode);
				reduced.sequence.addAll(0, hedge.sequence.subList(0, i));
				reduced.sequence.addAll(hedge.sequence.subList(i + 2,
						hedge.size()));
				return reduced.reduce(count - 1);
			}
		}
		return hedge;
	}

	@Override
	public void print(Writer out) throws IOException {
		print(out, true);
	}

	public void print(Writer out, boolean isBlock) throws IOException {
		print(out, isBlock, PRINT_PARAM_SEPARATOR, true, false);
	}

	private void print(Writer out, boolean isBlock, String separator,
			boolean needLamdaBracket, boolean friendlyOut) throws IOException {
		if (sequence.size() > 0) {
			if (isBlock)
				out.append(PRINT_PARAM_START);
			for (int i = 0, n = sequence.size(); i < n; i++) {
				if (i > 0)
					out.append(separator);
				if (friendlyOut)
					sequence.get(i).print(out, needLamdaBracket);
				else
					sequence.get(i).print(out);
			}
			if (isBlock)
				out.append(PRINT_PARAM_END);
		}
	}

	@Override
	public void print(Writer out, TermAtom atom, boolean needLamdaBracket)
			throws IOException {
		if (atom instanceof Lambda) {
			if (needLamdaBracket)
				out.append(PRINT_PARAM_START);
			atom.print(out);
			print(out, false, PRINT_PARAM_SEPARATOR, size() > 1
					|| needLamdaBracket, true);
			if (needLamdaBracket)
				out.append(PRINT_PARAM_END);
		} else if (atom instanceof Variable) {
			atom.print(out);
			print(out, true, Hedge.PRINT_PARAM_SEPARATOR, false, true);
		} else {
			atom.print(out);
			print(out, false, PRINT_PARAM_SEPARATOR, size() > 1
					|| needLamdaBracket, true);
		}
	}
}
