/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.data;

import java.util.Deque;
import java.util.Map;

import at.jku.risc.stout.hoau.data.atom.BoundVariable;
import at.jku.risc.stout.hoau.data.atom.Function;
import at.jku.risc.stout.hoau.data.atom.Lambda;
import at.jku.risc.stout.hoau.data.atom.TermAtom;
import at.jku.risc.stout.hoau.data.atom.Variable;
import at.jku.risc.stout.hoau.util.DataStructureFactory;

/**
 * A factory to create {@linkplain TermNode}s, {@linkplain Hedge}s and atomic
 * types ( {@linkplain TermAtom}s). This factory has also the ability to
 * automatically transform all the created bound variables so that they are all
 * guaranteed to be distinct.
 * 
 * @author Alexander Baumgartner
 */
public class NodeFactory {

	public static String PREFIX_IndividualVar;
	public static String PREFIX_Function;
	public static String PREFIX_Lambda;
	public static String PREFIX_Constant;
	public static String PREFIX_FreshFreeVar = "$";
	public static String SUFFIX_FreshFreeVar;
	public static String PREFIX_FreshBoundVar = "#";
	public static String SUFFIX_FreshBoundVar;

	public static boolean MAKE_BOUNDVAR_DISTINCT = true;

	private static long freshVarCnt = 0;
	private static long boundVarCnt = 0;

	private Deque<Hedge> hStack = DataStructureFactory.$.newDeque();
	private Deque<Lambda> lStack = DataStructureFactory.$.newDeque();
	private Map<String, Deque<BoundVariable>> boundMap = DataStructureFactory.$
			.newMap();

	/**
	 * Everything what is not an argument list of a static function symbol is a
	 * lambda hedge. We show an example where we denote lambda hedges with ()
	 * and simple function argument hedges with [] and with \ we denote the
	 * lambda symbol:<br>
	 * (\x.f[a, x(f, (b, c))]) \x.\y.g[x(y)]<br>
	 * This lambda term will reduce to f[a, g[f[b c]]]
	 */
	public void pushHedge(boolean lambda) {
		if (lambda)
			hStack.push(new LambdaHedge());
		else
			hStack.push(new Hedge());
	}

	public static void resetCounter() {
		freshVarCnt = 0;
		boundVarCnt = 0;
	}

	public void addToHedge(TermNode node) {
		hStack.peek().add(node);
	}

	public Hedge popHedge() {
		return hStack.pop();
	}

	public TermNode createFunction(String name, String type, Hedge hedge) {
		if (PREFIX_Function != null)
			name = PREFIX_Function + name;
		return new TermNode(new Function(name, type), hedge);
	}

	public TermNode createConstant(String name, String type) {
		if (PREFIX_Constant != null)
			name = PREFIX_Constant + name;
		return new TermNode(new Function(name, type), null);
	}

	public static Variable obtainFreshVar(String type) {
		StringBuilder name = new StringBuilder();
		if (PREFIX_FreshFreeVar != null)
			name.append(PREFIX_FreshFreeVar);
		name.append(++freshVarCnt);
		if (SUFFIX_FreshFreeVar != null)
			name.append(SUFFIX_FreshFreeVar);
		return new Variable(name.toString(), type);
	}

	public static BoundVariable obtainFreshBoundVar(String originName,
			String type) {
		StringBuilder name = new StringBuilder();
		if (PREFIX_FreshBoundVar != null)
			name.append(PREFIX_FreshBoundVar);
		name.append(++boundVarCnt); // uses same counter
		if (SUFFIX_FreshBoundVar != null)
			name.append(SUFFIX_FreshBoundVar);
		return new BoundVariable(name.toString(), originName, type);
	}

	public static TermNode obtainFreshVarNode(String type) {
		return new TermNode(obtainFreshVar(type), null);
	}

	public TermNode createIndividualVar(String name, String type) {
		return createIndividualVar(name, type, null);
	}

	public boolean hasBoundVar(String name) {
		return boundMap.containsKey(name);
	}

	public TermNode createIndividualVar(String name, String type, Hedge hedge) {
		Deque<BoundVariable> lambdaBound = boundMap.get(name);
		if (lambdaBound != null)
			return new TermNode(lambdaBound.peek(), hedge);
		if (PREFIX_IndividualVar != null)
			name = PREFIX_IndividualVar + name;
		return new TermNode(new Variable(name, type), hedge);
	}

	public void pushLambda(String boundVar, String type)
			throws MalformedTermException {
		if (boundVar.length() == 0)
			throw new MalformedTermException(
					"Lambda sign without bound variable");
		Lambda l;
		if (MAKE_BOUNDVAR_DISTINCT)
			l = new Lambda(obtainFreshBoundVar(boundVar, type));
		else
			l = new Lambda(new BoundVariable(boundVar, boundVar, type));
		String varName = l.getBoundVar().getOriginName();
		lStack.push(l);
		Deque<BoundVariable> boundVars = boundMap.get(varName);
		if (boundVars == null) {
			boundVars = DataStructureFactory.$.newDeque();
			boundMap.put(varName, boundVars);
		}
		boundVars.push(l.getBoundVar());
	}

	public TermNode popLambda(Hedge hedge) {
		Lambda l = lStack.pop();
		String varName = l.getBoundVar().getOriginName();
		Deque<BoundVariable> boundVars = boundMap.get(varName);
		boundVars.pop();
		if (boundVars.size() == 0)
			boundMap.remove(varName);
		return new TermNode(l, hedge);
	}
}
