/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.data;

import java.io.IOException;
import java.io.Writer;

import at.jku.risc.stout.hoau.data.atom.BoundVariable;
import at.jku.risc.stout.hoau.data.atom.Function;
import at.jku.risc.stout.hoau.data.atom.Lambda;
import at.jku.risc.stout.hoau.data.atom.TermAtom;
import at.jku.risc.stout.hoau.data.atom.Variable;
import at.jku.risc.stout.hoau.util.Printable;

/**
 * A term node consists of a {@linkplain TermAtom} and a {@linkplain Hedge}.
 * 
 * @author Alexander Baumgartner
 */
public class TermNode extends Printable implements Cloneable {
	private TermAtom atom;
	private Hedge hedge;
	/**
	 * The user friendly output form transforms lambda hedges to argument lists.
	 * For instance (\x.(\y. x y)) will be displayed as \x.\y. x(y)<br>
	 * Default = true
	 */
	public static boolean printUserFriendly = true;
	public static final TermAtom nullAtom = new TermAtom("") {
		@Override
		public void print(Writer out) throws IOException {
		}

		@Override
		public String getType() {
			return null;
		}

		@Override
		public String appendType(String type, char typeConstructor) {
			return type;
		}

		@Override
		public boolean isBasicType() {
			return false;
		}

		@Override
		public TermAtom clone() {
			return this;
		}
	};
	public static final Hedge nullHedge = new Hedge() {
		@Override
		public void add(TermNode t) {
			throw new IllegalAccessError("Illegal null-hedge modification");
		}

		@Override
		public Hedge reduce(int count) throws NotNormalizableException {
			return this;
		}

		@Override
		public Hedge clone() {
			return this;
		}
	};

	/**
	 * Instantiates a new term node. null values are transformed to unique
	 * objects which represent the null value. So that no
	 * {@linkplain NullPointerException} can occur.
	 * 
	 * @see TermNode#nullAtom
	 * @see TermNode#nullHedge
	 */
	public TermNode(TermAtom atom, Hedge hedge) {
		setAtom(atom);
		setHedge(hedge);
	}

	public TermAtom getAtom() {
		return atom;
	}

	/**
	 * Null values are transformed to unique objects which represent the null
	 * value. So that no {@linkplain NullPointerException} can occur.
	 * 
	 * @see TermNode#nullAtom
	 */
	public void setAtom(TermAtom atom) {
		this.atom = atom == null ? nullAtom : atom;
	}

	public Hedge getHedge() {
		return hedge;
	}

	/**
	 * Null values are transformed to unique objects which represent the null
	 * value. So that no {@linkplain NullPointerException} can occur.
	 * 
	 * @see TermNode#nullHedge
	 */
	public void setHedge(Hedge hedge) {
		this.hedge = hedge == null ? nullHedge : hedge;
	}

	/**
	 * This equality test performs an alpha equivalence test and also allows eta
	 * expansion in order to make the given {@linkplain TermNode}s equal.
	 */
	@Override
	public boolean equals(Object other) {
		if (!(other instanceof TermNode))
			return false;
		TermNode oNode = (TermNode) other;
		if (atom.equals(oNode.atom)) {
			Hedge oHedge = oNode.hedge;
			// Alpha equivalence
			if (atom instanceof Lambda) {
				Lambda thisA = (Lambda) atom;
				Lambda otherA = (Lambda) oNode.atom;
				if (!otherA.getBoundVar().equals(thisA.getBoundVar())) {
					oNode.atom = atom;
					oHedge.substitute(otherA.getBoundVar(), thisA.getBoundVar());
				}
			}
			return hedge.equals(oHedge);
		}
		// Try eta-expansion
		else if (atom instanceof Lambda) {
			if (oNode.atom instanceof Lambda)
				return false; // Bound variable type mismatch
			oNode.etaExpand(((Lambda) this.atom).getBoundVar());
			return this.equals(oNode);
		} else if (oNode.atom instanceof Lambda) {
			this.etaExpand(((Lambda) oNode.atom).getBoundVar());
			return this.equals(oNode);
		}
		return false;
	}

	public void etaExpand(BoundVariable boundVar) {
		boundVar = NodeFactory.obtainFreshBoundVar(boundVar.getOriginName(),
				boundVar.getType());
		if (hedge == nullHedge)
			hedge = atom instanceof Function ? new Hedge() : new LambdaHedge();
		hedge.add(new TermNode(boundVar, null));
		if (!isNullAtom()) {
			LambdaHedge newH = new LambdaHedge();
			newH.add(new TermNode(atom, hedge));
			hedge = newH;
		}
		atom = new Lambda(boundVar);
	}

	/**
	 * Substitutes a variable by a term and returns the result.
	 */
	public void substitute(Variable fromVar, TermNode toTerm)
			throws NotNormalizableException {
		if (fromVar.equals(atom)) {
			if (hedge.isEmpty()) {
				hedge = toTerm.hedge.clone();
				atom = toTerm.atom.clone();
			} else if (toTerm.hedge.isEmpty()) {
				substHedge(fromVar, toTerm, toTerm.hedge.newInstance());
				atom = toTerm.atom.clone();
			} else if (toTerm.atom instanceof Function) {
				substHedge(fromVar, toTerm, toTerm.hedge.clone());
				atom = toTerm.atom.clone();
			} else {
				LambdaHedge args = new LambdaHedge();
				args.add(toTerm.clone());
				substHedge(fromVar, toTerm, args);
				atom = nullAtom;
			}
		} else if (atom instanceof Lambda) {
			BoundVariable lBound = ((Lambda) atom).getBoundVar();
			if (!lBound.equals(fromVar)) {
				atom = new Lambda(NodeFactory.obtainFreshBoundVar(
						lBound.getOriginName(), lBound.getType()));
				hedge.substitute(lBound, ((Lambda) atom).getBoundVar());
				hedge.substitute(fromVar, toTerm);
			}
		} else {
			hedge.substitute(fromVar, toTerm);
		}
	}

	private void substHedge(Variable fromVar, TermNode toTerm, Hedge args)
			throws NotNormalizableException {
		hedge.substitute(fromVar, toTerm);
		args.sequence.addAll(hedge.sequence);
		hedge = args;
	}

	/**
	 * Substitutes one bound variable by another bound variable.
	 */
	public void substitute(BoundVariable from, BoundVariable to) {
		if (from.equals(atom))
			atom = to;
		if (!(atom instanceof Lambda && ((Lambda) atom).getBoundVar().equals(
				from)))
			hedge.substitute(from, to);
	}

	@Override
	public TermNode clone() {
		return new TermNode(atom.clone(), hedge.clone());
	}

	@Override
	public void print(Writer out) throws IOException {
		if (printUserFriendly) {
			print(out, hedge.size() > 1);
		} else {
			if (hedge instanceof LambdaHedge && atom != nullAtom) {
				out.append(LambdaHedge.PRINT_PARAM_START);
				atom.print(out);
				out.append(LambdaHedge.PRINT_PARAM_SEPARATOR);
				((LambdaHedge) hedge).print(out, false);
				out.append(LambdaHedge.PRINT_PARAM_END);
			} else {
				atom.print(out);
				hedge.print(out);
			}
		}
	}

	void print(Writer out, boolean needLambdaBracket) throws IOException {
		hedge.print(out, atom, needLambdaBracket);
	}

	/**
	 * Tests whether the atom of this term node is the null atom.
	 */
	public boolean isNullAtom() {
		return atom == nullAtom;
	}

	/**
	 * Perform beta reduction with the given upper bound for recursion
	 * repetition to avoid endless recursion in untyped calculus.
	 */
	public TermNode reduce(int maxSteps) throws NotNormalizableException {
		hedge = hedge.reduce(maxSteps);
		if (isNullAtom() && hedge.size() == 1)
			return hedge.get(0);
		return this;
	}

	/**
	 * Reduce unnecessary leading hedges. For instance ((f(x))) will be reduced
	 * to f(x).
	 */
	public TermNode reduceHedges() {
		if (isNullAtom() && hedge.size() == 1)
			return hedge.get(0).reduceHedges();
		return this;
	}

	// This more elegant closure implementation would be slower :/
	// public boolean hasFreeVar(final BoundVariable var) {
	// return traverse(new TraverseCallBack() {
	// public boolean exec(TermAtom atom, Hedge hedge) {
	// return var.equals(atom);
	// }});}
	/**
	 * Checks whether a variable occurs freely inside inside the term.
	 */
	public boolean hasFreeVar(Variable var) {
		if (var.equals(getAtom()))
			return true;
		for (int i = 0, n = hedge.size(); i < n; i++)
			if (hedge.get(i).hasFreeVar(var))
				return true;
		return false;
	}

	/**
	 * Traverses the term tree and executes the callback function at every node.
	 * 
	 * @see TraverseCallBack
	 */
	public boolean traverse(TraverseCallBack callBack) {
		if (callBack.exec(atom, hedge))
			return true;
		for (int i = 0, n = hedge.size(); i < n; i++)
			if (hedge.get(i).traverse(callBack))
				return true;
		return false;
	}

	/**
	 * This class may be used to
	 * {@linkplain TermNode#traverse(TraverseCallBack) traverse} a term tree and
	 * execute an arbitrary operation on every node.
	 * 
	 * @author Alexander Baumgartner
	 */
	public abstract class TraverseCallBack {
		/**
		 * Propagation will stop as soon as this method returns true.
		 */
		public abstract boolean exec(TermAtom atom, Hedge hedge);
	}
}
